#!/usr/bin/env python
import sys, os, re, random, math, subprocess, time, platform;
from operator import itemgetter;

def safeprint(*args):
	print(" ".join([str(arg) for arg in args]))

def labelsortfunc(x):
	return x[1][1]

class ConversionComplete(Exception):
     pass

def disphelp():
	safeprint("\n BranchMap Help Menu:\n======================\n",\
		"branchmap.py [-hcbpsrta] [asmfile]\n",\
		"-h:     Display this help menu\n",\
		"-c:     Show calls in map file\n",\
		"-b:     Show bcalls in map file\n",\
		"-p:     Use ps2pdf to convert PS to PDF\n",\
		"-s:     Show push/pop stack manipulation\n",\
		"-r:     Display breaks between routines (best-guess)\n",\
		"-t:     Show tabguides\n",\
		"-a:     Verbose output, equivalent to \"-c -b -p -s -r -t\"\n",\
		"asmfile: A z80 assembly-language file to parse\n")
	sys.exit(0)
	return

t_start = time.localtime()
g_labelcount = 0
g_branchcount = 0
g_callcount = 0
g_bcallcount = 0
g_numfiles = 0
g_filelist = {}
g_labellist = {}
g_branchlist = {}
g_instlist = []				#stores instructions: 0=jumps, 1=calls, 2=bcalls, 3=implied jumps, 4=rets, 5=pushes, 6=pops
g_warnlist = []
g_defines = []
g_showcalls = False
g_showbcalls = False
g_genpdf = False
g_showpushpop = False
g_showbreaks = False
g_tabguides = False

def parsefile(filename,parent,parentline,abspath,mode):
	global g_labelcount, g_branchcount, g_callcount, g_bcallcount, g_numfiles
	global g_filelist, g_labellist, g_branchlist, g_instlist, g_warnlist, g_defines
	global g_showcalls, g_showbcalls

	context = "**begin**"
	thisfilenum = g_numfiles
	g_numfiles = g_numfiles+1
	while True:
		g_filelist[thisfilenum] = filename, (random.random(), random.random(), random.random())
		if 1.5 > sum(g_filelist[thisfilenum][1]):
			break

	if mode == 1:
		safeprint("Searching through " + filename)
	if mode == 2:
		safeprint("Verifying " + filename)
	fullpath = os.path.join(abspath,filename)
	if not(os.path.exists(fullpath)):
		safeprint("Error: "+fullpath+" was not found!")
		return
	fh = open(fullpath,'r')
	labelcount = 0
	branchcount = 0
	callcount = 0
	bcallcount = 0
	linenum = 1
	stackdepth = 0
	undefined = 0
	execpath = True			#when set to false, the current line is unreachable from normal, non-fancy program flow
	#cmdsstart = False

	fwsplit = re.compile(r'[^a-zA-Z0-9_#\.]')
	g_labellist[("*"+filename+"start").lower()] = thisfilenum, 0, ("*"+filename+"start"), -999

	for line in fh:
	
		while line:
			if line[:len(";branchmap_root ")] == ";branchmap_root ":
				abspath = os.path.normpath(abspath + "/" + line[len(";branchmap_root "):-1])
			if line.find(";") != -1:
				line = line[:line.find(";")]
			
			linestrip = line.strip()
			linelow = linestrip.lower()
			firstword = fwsplit.split(linelow)[0]

			#safeprint(filename," ",linenum,": [",line,"]")
			#if len(line)<2 and len(line) != 0:
			#	sys.exit(0)

			if len(linelow) > len(firstword) and linelow[len(firstword)] == ":" and len(linelow) > len(firstword)+1:
				line = linestrip[len(firstword)+1:]
				linestrip = linestrip[:len(firstword)+1].strip()
				linelow = linestrip.lower()
				#safeprint("BLAH1 [",linestrip,"] [",line,"]")
			elif linelow.find("\\") != -1:
				line = linestrip[linelow.find("\\")+1:]
				linestrip = linestrip[:linelow.find("\\")-1].strip()
				linelow = linestrip.lower()
				#safeprint("BLAH2 [",linestrip,"] [",line,"]")
			else:
				line = ""
				#safeprint("BLAH3 [",linestrip,"]")

			if (not(undefined)):
				if firstword == "#include":
					fname = linelow[len("#include")+2:-1]
					if execpath:
						g_instlist.append((thisfilenum, linenum, context, "*"+fname+"start", 4, stackdepth))
					parsefile(fname, thisfilenum, linenum, abspath, mode)
				elif firstword == "jr" or firstword == "jp" or firstword == "djnz":
					branchcount = branchcount+1
					jumplabel = fwsplit.split(linestrip[len(firstword)+1:].strip())
					#safeprint("+ [" + jumplabel[-1] + "]")
					g_instlist.append((thisfilenum, linenum, context, jumplabel[-1], 0, stackdepth))	 #0 indicates jump in jumplist
					if "," not in linestrip[len(firstword)+1:] and firstword != "djnz":
						execpath = False
						context = ""
						if g_showbreaks:
							g_instlist.append((thisfilenum, linenum, context, "", 7, stackdepth))
						stackdepth = 0
				elif firstword == "call":
					callcount = callcount+1
					jumplabel = fwsplit.split(linestrip[len(firstword)+1:].strip())
					g_instlist.append((thisfilenum, linenum, context, jumplabel[-1], 1, stackdepth)) #1 indicates call in jumplist
				elif firstword == "bcall" or firstword == "bcallz" or \
					firstword == "bcallnz" or firstword == "bcallc" or \
					firstword == "bcallnc":
					bcallcount = bcallcount+1
					if linestrip.find("("):
						bcalllabel = linestrip[linestrip.find("(")+1:linestrip.find(")")]
					else:
						bcalllabel = (fwsplit.split(linestrip[len(firstword)+1:].strip()))[-1]
					g_instlist.append((thisfilenum, linenum, context, bcalllabel[1:], 2, stackdepth)) #2 indicates bcall in jumplist
				elif firstword == "b_call" or firstword == "b_callz" or \
					firstword == "b_callnz" or firstword == "b_callc" or \
					firstword == "b_callnc":
					bcallcount = bcallcount+1
					if linestrip.find("("):
						bcalllabel = linestrip[linestrip.find("(")+1:linestrip.find(")")]
					else:
						bcalllabel = (fwsplit.split(linestrip[len(firstword)+1:].strip()))[-1]
					g_instlist.append((thisfilenum, linenum, context, bcalllabel[1:], 2, stackdepth)) #2 indicates bcall in jumplist
				elif firstword == "push":
					if g_showpushpop:
						g_instlist.append((thisfilenum, linenum, context, "", 5, stackdepth))
					stackdepth += 1
				elif firstword == "pop":
					if g_showpushpop:
						g_instlist.append((thisfilenum, linenum, context, "", 6, stackdepth))
					stackdepth -= 1
				elif firstword == "ret":
					if len(linelow) == len(firstword):		#ie, not a ret z / ret nz, etc, so no execution can flow though
						g_instlist.append((thisfilenum, linenum, context, linelow[len(firstword)+1:], 4, stackdepth))
						execpath = False
						context = ""
						if g_showbreaks:
							g_instlist.append((thisfilenum, linenum, context, "", 7, stackdepth))
						stackdepth = 0
				elif linelow[-1:] == ":":
					labelcount += 1
					if execpath == True:
						g_instlist.append((thisfilenum, linenum, context, linestrip[:-1], 3, stackdepth))	 #3 indicates implicit jump/continue
					g_labellist[linestrip[:-1].lower()] = thisfilenum, linenum, linestrip[:-1], -999

					stackdepth = 0
					context = linestrip[:-1]
					execpath = True
				
			if firstword == "#define":
				g_defines.append(linestrip[len(firstword)+1:])
			elif firstword == "#ifdef":
				if not(linestrip[len(firstword)+1:] in g_defines):
					undefined = 1
			elif firstword == "#endif":
				undefined = 0
			elif firstword == "#else":
				undefined = not(undefined)

		linenum = linenum + 1

	g_branchcount = g_branchcount + branchcount
	g_callcount = g_callcount + callcount
	g_bcallcount = g_bcallcount + bcallcount
	g_labelcount = g_labelcount + labelcount
	return

def analyzeblock(label,initdepth):
	global g_labelcount, g_branchcount, g_callcount, g_bcallcount, g_numfiles
	global g_filelist, g_labellist, g_branchlist, g_instlist, g_warnlist, g_defines
	global g_showcalls, g_showbcalls

	adjusted = 0
	complete = False
	
	while not(complete):
		complete = True
		labellower = label.lower()
		if labellower not in g_labellist:
			break
		if g_labellist[labellower][3] == -999:
			#safeprint(">>> In on label "+label+" at depth "+str(initdepth))
			#safeprint("Adjusting " + labellower + " from " + str(g_labellist[labellower][3]) + " to " + str(initdepth))
			g_labellist[labellower] = g_labellist[labellower][:-1] + (initdepth,)
			adjusted += 1
			for i in range(0,len(g_instlist)):
				jump = g_instlist[i]
				#safeprint(i,jump,)
				if jump[2].lower() == labellower:
					#safeprint("Adjusting jump/call (" + str(i) + ") " + jump[3] + " from " + str(jump[5]) + " to " + str(initdepth+jump[5]))
					g_instlist[i] = jump[:-1] + ((initdepth+jump[5]),)
					#print g_instlist[i]
					#if g_labellist[jump[3].lower()][3] == -999:
					#	g_labellist[jump[3].lower()] = g_labellist[jump[3].lower()][:-1] + ((initdepth+jump[5]),)
					jump = g_instlist[i]
					#safeprint(jump)
					adjusted += 1
					if jump[4] == 4:
						if jump[5] != 0:
							g_warnlist.append((jump[0], jump[1], jump[2], jump[5]))							
					elif jump[4] < 4 and jump[3].lower() in g_labellist and g_labellist[jump[3].lower()][3] == -999:
						if jump[4] == 0:
							if labellower == jump[3].lower():
								safeprint("Unexpected recursion detected on",labellower,", aborting now")
							else:
								adjusted = adjusted + analyzeblock(jump[3],jump[5])
								#safeprint("<<< Back in label "+label+" at depth "+str(initdepth))
						elif jump[4] == 3:
							label = jump[3]
							initdepth = jump[5]
							complete = False
							break
		
	return adjusted

safeprint(" --+=====================+--")
safeprint(" --|   BranchMap v1.0    |--")
safeprint(" --|   by Kerm Martian   |--")
safeprint(" --|   www.Cemetech.net  |--")
safeprint(" --|  admin@cemetech.net |--")
safeprint(" --+=====================+--")
rootfile = ""
if (len(sys.argv) > 1):
	prefix = ""
	for i in range(1,len(sys.argv)):
		if sys.argv[i][0] == "-":
			if sys.argv[i][1] == "h":
				disphelp()
			elif sys.argv[i][1] == "c":
				g_showcalls = True
			elif sys.argv[i][1] == "b":
				g_showbcalls = True
			elif sys.argv[i][1] == "p":
				g_genpdf = True
			elif sys.argv[i][1] == "s":
				g_showpushpop = True
			elif sys.argv[i][1] == "r":
				g_showbreaks = True
			elif sys.argv[i][1] == "t":
				g_tabguides = True
			elif sys.argv[i][1] == "a":
				g_showcalls = True
				g_showbcalls = True
				g_genpdf = True
				g_showpushpop = True
				g_showbreaks = True
				g_tabguides = True
			else:
				safeprint("\nUnknown switch: ",sys.argv[i])
				disphelp()
			#prefix = sys.argv[i][1]
		else:
			if prefix == "":
				rootfile = sys.argv[i]
			else:
				safeprint("\nUnknown switch: -",prefix,"\n")

if rootfile == "":
	safeprint("\nPlease enter the relative path to the z80 ASM file to analyze: ")
	rootfile = sys.stdin.readline()[:-1]

safeprint("**Pass 1: Control Structure Extraction**")
basepath = os.path.dirname(rootfile)
rootfile = os.path.basename(rootfile)
parsefile(rootfile,-1,-1,basepath,1)		#-1 = this is the root parent, [rootfile] has no parent
safeprint("-------------------------------")
safeprint("**Pass 1**: Final Summary:\n   " + str(g_branchcount) + " branches\n   " + str(g_callcount) + \
	" calls\n   " + str(g_bcallcount) + " bcalls\n   " + str(g_labelcount) + " labels")

safeprint("\n**Pass 2: Stack Depth Analysis**")

g_instlist.sort(key=itemgetter(0,1,4))				#this is quite important!
labels = list(g_labellist.items())	
labels.sort(key=labelsortfunc)

adjusted = 0
for jump in g_instlist:
	if jump[4] == 1 or jump[4] == 2:
		adjusted += analyzeblock(jump[3],0)

for jump in g_instlist:
	if jump[1] == 0:
		adjusted += analyzeblock(jump[3],0)

#safeprint("\n-=-=-=-=-=-=\n")
for label in labels:
	if g_labellist[label[0]][3] == -999:
		adjusted += analyzeblock(label[1][2],0)
		
safeprint("**Pass 2**: Final Summary:\n   " + str(adjusted) + " labels/instructions adjusted\n   " + str(len(g_warnlist)) + \
	" stack manipulation warnings")

labels = list(g_labellist.items())	
labels.sort(key=labelsortfunc)

safeprint("\n**Pass 3: Postscript Generation**")
psname = os.path.join(basepath,rootfile[:-1*(rootfile[::-1].find(".")+1)]+".ps")
ps = open(psname,'w')
#letter height = 792pt, width = 612pt
margin_t = 760;
margin_b = 40;
margin_l = 32;
margin_r = 580;
margin_tab = 30;
ps.write("%!\n\n/Courier findfont\n10 scalefont\nsetfont\n\n");

page_y = margin_t
ps.write("gsave\n/Courier findfont\n20 scalefont\nsetfont\n" + str(margin_l) + " " + \
	str(page_y) + " moveto\n(BranchMap Summary File for ")
ps.write(rootfile + ") show\ngrestore\n")
page_y -= 72
ps.write("gsave\n/Courier findfont\n14 scalefont\nsetfont\n" + str(margin_l) + " " + \
	str(page_y) + " moveto\n(ASM File List:) show\n")
ps.write("grestore\n")
page_y -= 16
for filenum in g_filelist:
	ps.write(str(g_filelist[filenum][1][0]) + " ")
	ps.write(str(g_filelist[filenum][1][1]) + " ")
	ps.write(str(g_filelist[filenum][1][2]) + " setrgbcolor\n")
	ps.write("gsave\n/Courier findfont\n10 scalefont\nsetfont\n" + str(margin_l) + " " + str(page_y) + " moveto\n(")
	ps.write(str(filenum+1) + "\) " + g_filelist[filenum][0] + ") show\ngrestore\n")
	page_y -= 12
	if page_y <= margin_b:
		ps.write("showpage\n\n")
		page_y = margin_t
		
ps.write("showpage\n\n")
page_y = margin_t
ps.write(str(margin_l) + " " + str(page_y) + " moveto\n(Project Statistics) show\n")
ps.write(str(margin_l) + " " + str(page_y-12) + " moveto\n(Total branches: "+ str(g_branchcount) + ") show\n")
ps.write(str(margin_l) + " " + str(page_y-24) + " moveto\n(Total calls: "+ str(g_callcount) + ") show\n")
ps.write(str(margin_l) + " " + str(page_y-36) + " moveto\n(Total bcalls: "+ str(g_bcallcount) + ") show\n")
ps.write(str(margin_l) + " " + str(page_y-48) + " moveto\n(Total labels: "+ str(g_labelcount) + ") show\n")
ps.write(str(margin_l) + " " + str(page_y-60) + " moveto\n(Stack Manipulation Warnings: "+ str(len(g_warnlist)) + ") show\n")
page_y = margin_t-80

g_warnlist.sort(key=itemgetter(0,1))
ps.write(str(margin_l) + " " + str(page_y) + " moveto\n(Stack Manipulation Warnings:) show\n")
page_y -= 10
for warning in g_warnlist:
	ps.write(str(margin_l) + " " + str(page_y) + " moveto\n(" + g_filelist[warning[0]][0] + ":" + str(warning[1]) + " in label " + warning[2])
	ps.write(": return with " + str(int(math.fabs(warning[3]))) + " too " + "many"*int(warning[3] < 0) + "few"*int(warning[3] > 0))
	ps.write(" pops) show\n")
	page_y -= 10
	if page_y <= margin_b:
		ps.write("showpage\n\n")
		page_y = margin_t

if page_y != margin_t:
	ps.write("showpage\n\n")

instskipped = 0
jumpindex = 0
for filenum in g_filelist:
	page_y = margin_t
	ps.write(str(g_filelist[filenum][1][0]) + " ")
	ps.write(str(g_filelist[filenum][1][1]) + " ")
	ps.write(str(g_filelist[filenum][1][2]) + " setrgbcolor\n")
	ps.write("gsave\n/Courier findfont\n14 scalefont\nsetfont\n" + str(margin_l) + " " + str(page_y) + " moveto\n(")
	ps.write(g_filelist[filenum][0] + ") show\ngrestore\n")
	page_y = page_y - 20
	
	labelindex = 0
	while 1:
		if labelindex >= len(labels):
			break
		labelindex += 1
		if (labels[labelindex-1][1][0] != filenum):
			continue


		# jumplist = []
		# for jumpitem in g_instlist:
			# if jumpitem[3].lower() == labels[labelindex-1][0]:
				# jumplist.append(jumpitem)
		
		if labels[labelindex-1][1][1] > 0:
			ps.write(str(margin_l) + " " + str(page_y) + " moveto\n(" + \
				" "*int(4-math.floor(math.log10(labels[labelindex-1][1][1]))) + \
				"[" + str(labels[labelindex-1][1][1]) + "] " + \
				str(labels[labelindex-1][1][2]) + ":) show\n")
			page_y  = page_y - 11
			if page_y <= margin_b:
				ps.write("showpage\n\n")
				ps.write(str(g_filelist[filenum][1][0]) + " ")
				ps.write(str(g_filelist[filenum][1][1]) + " ")
				ps.write(str(g_filelist[filenum][1][2]) + " setrgbcolor\n")
				page_y = margin_t
		
		while 1:
			#safeprint(jumpindex,g_instlist[jumpindex],)
			while jumpindex < len(g_instlist) and g_instlist[jumpindex][0] < filenum:
				jumpindex+=1
			if jumpindex >= len(g_instlist) or g_instlist[jumpindex][0] != filenum:
				break

			while jumpindex < len(g_instlist) and g_instlist[jumpindex][0] == filenum and g_instlist[jumpindex][1] <= labels[labelindex-1][1][1] and g_instlist[jumpindex][2].lower() != labels[labelindex-1][0]:
				if g_instlist[jumpindex][4] == 0:
					instskipped += 1
				jumpindex+=1
			if jumpindex >= len(g_instlist) or g_instlist[jumpindex][2].lower() != labels[labelindex-1][0]:
				break
			jumpindex += 1
			type = g_instlist[jumpindex-1][4]
			if type == 7 or type == 4:
				if not(g_showbreaks):
					continue
				ps.write("newpath\n" + str(margin_l) + " " + str(page_y+7) + " moveto\n")
				ps.write(str(margin_r) + " " + str(page_y+7) + " lineto\nstroke\n")
				page_y -= 3
				if page_y <= margin_b:
					ps.write("showpage\n\n")
					ps.write(str(g_filelist[filenum][1][0]) + " ")
					ps.write(str(g_filelist[filenum][1][1]) + " ")
					ps.write(str(g_filelist[filenum][1][2]) + " setrgbcolor\n")
					page_y = margin_t
				continue
			if g_showpushpop and type >= 5:			#push/pop
				bottomleft = margin_l+margin_tab*g_instlist[jumpindex-1][5]+78
				if type == 5:
					ps.write("newpath\n" + str(bottomleft) + " " + str(page_y+6) + " moveto\n")
					ps.write(str(bottomleft+margin_tab) + " " + str(page_y+6) + " lineto\n")
					ps.write(str(bottomleft+margin_tab-2) + " " + str(page_y+4) + " lineto\n")
					ps.write(str(bottomleft+margin_tab) + " " + str(page_y+6) + " lineto\n")
					ps.write(str(bottomleft+margin_tab-2) + " " + str(page_y+8) + " lineto\nstroke\n")
				else:
					ps.write("newpath\n" + str(bottomleft) + " " + str(page_y+6) + " moveto\n")
					ps.write(str(bottomleft-margin_tab) + " " + str(page_y+6) + " lineto\n")
					ps.write(str(bottomleft-margin_tab+2) + " " + str(page_y+4) + " lineto\n")
					ps.write(str(bottomleft-margin_tab) + " " + str(page_y+6) + " lineto\n")
					ps.write(str(bottomleft-margin_tab+2) + " " + str(page_y+8) + " lineto\nstroke\n")
				page_y -= 4
				if page_y <= margin_b:
					ps.write("showpage\n\n")
					ps.write(str(g_filelist[filenum][1][0]) + " ")
					ps.write(str(g_filelist[filenum][1][1]) + " ")
					ps.write(str(g_filelist[filenum][1][2]) + " setrgbcolor\n")
					page_y = margin_t
				continue
			if (type == 1 and not(g_showcalls)) or (type == 2 and not(g_showbcalls)) or \
				type >= 3:
				continue
			tabs = g_instlist[jumpindex-1][5]
			if tabs < 0:
				tabs = 0
			if g_tabguides:
				for i in range(0,tabs):
					ps.write("newpath\n"+str(margin_l+78+30*i)+" "+str(page_y+9)+" moveto\n"+\
						str(margin_l+78+30*i)+" "+str(page_y)+" lineto\n[1 2] 0 setdash\nstroke\n[] 0 setdash\n")
			ichar = " cbcppj"[type]
			ps.write(str(margin_l) + " " + \
				str(page_y) + " moveto\n(" + ichar + \
				" "*int(3-math.floor(math.log10(g_instlist[jumpindex-1][1]))) + "[" + \
				str(g_instlist[jumpindex-1][1]) + \
				"] "+" "*int(5*(tabs+1)))
			bottomleft = margin_l-3+6*(9+len(g_instlist[jumpindex-1][3]))+margin_tab*(tabs+1)
			if g_instlist[jumpindex-1][5] < 0:
				ps.write(") show\ngsave\n1.0 0.0 0.0 setrgbcolor\n([!] ) show\ngrestore\n(    ")
				bottomleft += 24
			try:
				if g_labellist[g_instlist[jumpindex-1][3].lower()][0] == filenum:
					if (type == 2):
						ps.write("_")
						bottomleft += 6				#the width of a Courier, 10-pt char
					ps.write(g_instlist[jumpindex-1][3] + ") show\n")
					if g_labellist[g_instlist[jumpindex-1][3].lower()][1] < g_instlist[jumpindex-1][1]:
						#upward arrow
						ps.write("newpath\n" + str(bottomleft) + " " + str(page_y) + " moveto\n")
						ps.write(str(bottomleft+4) + " " + str(page_y+8) + " lineto\n")
						ps.write(str(bottomleft+8) + " " + str(page_y) + " lineto\nclosepath\nfill\nstroke\n")
					else:
						#downward arrow
						ps.write("newpath\n" + str(bottomleft) + " " + str(page_y+8) + " moveto\n")
						ps.write(str(bottomleft+4) + " " + str(page_y) + " lineto\n")
						ps.write(str(bottomleft+8) + " " + str(page_y+8) + " lineto\nclosepath\nfill\nstroke\n")
				else:
					#pointer to other page
					ps.write(") show\ngsave\n")
					ps.write(str(g_filelist[g_labellist[g_instlist[jumpindex-1][3].lower()][0]][1][0]) + " ")
					ps.write(str(g_filelist[g_labellist[g_instlist[jumpindex-1][3].lower()][0]][1][1]) + " ")
					ps.write(str(g_filelist[g_labellist[g_instlist[jumpindex-1][3].lower()][0]][1][2]) + " setrgbcolor\n(")
					if (type == 2):
						ps.write("_")
						bottomleft += 6				#the width of a Courier, 10-pt char
					ps.write(g_instlist[jumpindex-1][3] + ") show\n")
					ps.write("newpath\n" + str(bottomleft) + " " + str(page_y+4) + " moveto\n")
					ps.write(str(margin_r-margin_tab) + " " + str(page_y+4) + " lineto\n")
					ps.write(str(margin_r-margin_tab-4) + " " + str(page_y) + " lineto\n")
					ps.write(str(margin_r-margin_tab) + " " + str(page_y+4) + " lineto\n")
					ps.write(str(margin_r-margin_tab-4) + " " + str(page_y+8) + " lineto\nstroke\n")
					ps.write(str(margin_r-margin_tab-10-6*(len(g_filelist[g_labellist[g_instlist[jumpindex-1][3].lower()][0]][0]))) + " " + \
						str(page_y+6) + " moveto\n(" + g_filelist[g_labellist[g_instlist[jumpindex-1][3].lower()][0]][0] + ") show\n")
					ps.write("grestore\n")
					page_y = page_y
			except KeyError:
				if (type == 2):
					ps.write("_")
				ps.write(g_instlist[jumpindex-1][3] + " ?????) show\n")
			page_y  -= 11
			
		if page_y <= margin_b:
			ps.write("showpage\n\n")
			ps.write(str(g_filelist[filenum][1][0]) + " ")
			ps.write(str(g_filelist[filenum][1][1]) + " ")
			ps.write(str(g_filelist[filenum][1][2]) + " setrgbcolor\n")
			page_y = margin_t
			
	if page_y != margin_t:
		ps.write("showpage\n\n")

ps.close()
safeprint("**Pass 3**: Completed.  Output written to "+psname)
if instskipped > 0:
	safeprint("   "+str(instskipped)+" jumps/instructions were not rendered.  This is likely because "+\
		"they are part of a jump table; BranchMap is unable to correctly determine flow around "+\
		"memory-indexed constructions such as jump tables, so it has likely discarded what it "+\
		"thinks are unreachable jumps.  It may have legitimately discarded unreachable code, however.")
if g_genpdf:
	safeprint("\n**Pass 4: Converting to PDF**")
	pathsep = ":"
	if (platform.system() == 'Windows'):
		pathsep = ";"
	try:
		for pathcomp in os.environ['PATH'].split(pathsep):
			for root, dirs, files in os.walk(pathcomp):
				if os.path.exists(os.path.join(root,"ps2pdf.exe")) or os.path.exists(os.path.join(root,"ps2pdf")):
					safeprint(os.path.join(root,"ps2pdf")+" \""+psname+"\" \""+psname[:-2]+"pdf\"")
					retcode = subprocess.call(["ps2pdf",psname,psname[:-2]+"pdf"])
					converted = True
					raise ConversionComplete
		safeprint("Error: ps2pdf could not be found.  Make sure ghostscript is installed.")
	except ConversionComplete:
		pass
	safeprint("\n**Pass 4**: Complete.")

safeprint("Total time:",str(time.mktime(time.localtime())-time.mktime(t_start)) + "s")
sys.exit()
